#!/usr/bin/python
# -*- coding: utf-8 -*-
# Taken from: https://github.com/solominh/python-stardict
import struct
import gzip


class IfoFileException(Exception):
    """Exception while parsing the .ifo file.
    Now version error in .ifo file is the only case raising this exception.
    """

    def __init__(self, description="IfoFileException raised"):
        """Constructor from a description string.

        Arguments:
        - `description`: a string describing the exception condition.
        """
        self._description = description

    def __str__(self):
        """__str__ method, return the description of exception occured.

        """
        return self._description


class IfoFileReader(object):
    """Read infomation from .ifo file and parse the infomation a dictionary.
    The structure of the dictionary is shown below:
    {key, value}
    """

    def __init__(self, filename):
        """Constructor from filename.

        Arguments:
        - `filename`: the filename of .ifo file of stardict.
        May raise IfoFileException during initialization.
        """
        self._ifo = dict()
        with open(filename, "rU", encoding='utf-8') as ifo_file:
            self._ifo["dict_title"] = ifo_file.readline()  # dictionary title
            line = ifo_file.readline()  # version info
            key, equal, value = line.partition("=")
            key = key.strip()
            value = value.strip()
            # check version info, raise an IfoFileException if error encounted
            if key != "version":
                raise IfoFileException(
                    "Version info expected in the second line of {!r:s}!".format(filename))
            if value != "2.4.2" and value != "3.0.0":
                raise IfoFileException(
                    "Version expected to be either 2.4.2 or 3.0.0, but {!r:s} read!".format(value))
            self._ifo[key] = value
            # read in other infomation in the file
            for line in ifo_file:
                key, equal, value = line.partition("=")
                key = key.strip()
                value = value.strip()
                self._ifo[key] = value
            # check if idxoffsetbits should be discarded due to version info
            if self._ifo["version"] == "3.0.0" and "idxoffsetbits" in self._ifo:
                del self._ifo["version"]

    def get_ifo(self, key):
        """Get configuration value.

        Arguments:
        - `key`: configuration option name
        Return:
        - configuration value corresponding to the specified key if exists, otherwise False.
        """
        if key not in self._ifo:
            return False
        return self._ifo[key]


class IdxFileReader(object):
    """Read dictionary indexes from the .idx file and store the indexes in a list and a dictionary.
    The list contains each entry in the .idx file, with subscript indicating the entry's origin index in .idx file.
    The dictionary is indexed by word name, and the value is an integer or a list of integers pointing to
    the entry in the list.
    """

    def __init__(self, filename, compressed=False, index_offset_bits=32):
        """
        Note: filename.dict and filename.dict.dz have different indices

        Arguments:
        - `filename`: the filename of .idx file of stardict.
        - `compressed`: indicate whether the .idx file is compressed.
        - `index_offset_bits`: the offset field length in bits.
        """
        # Open file
        if compressed:
            with gzip.open(filename, "rb") as index_file:
                self._content = index_file.read()
        else:
            with open(filename, "rb") as index_file:
                self._content = index_file.read()

        # Init
        self._offset = 0
        self._index = 0
        self._index_offset_bits = index_offset_bits
        self._word_idx = {}
        self._index_idx = []

        # Indexing
        for word_str, word_data_offset, word_data_size, index in self:  # Call __iter__ function
            self._index_idx.append(
                (word_str, word_data_offset, word_data_size))
            if word_str in self._word_idx:
                self._word_idx[word_str].append(len(self._index_idx) - 1)
            else:
                self._word_idx[word_str] = [len(self._index_idx) - 1]

        # Clean up
        del self._content
        del self._index_offset_bits
        del self._index

    def __iter__(self):
        """Define the iterator interface.

        """
        return self

    def __next__(self):
        """Define the iterator interface.

        """
        # EOF
        if self._offset == len(self._content):
            raise StopIteration
        # Read word_str => end with \0
        end = self._content.find(b'\x00', self._offset)
        word_str = self._content[self._offset: end].decode('utf-8')
        self._offset = end + 1

        # Read word_data_offset => 32 or 64 bits
        word_data_offset = 0
        if self._index_offset_bits == 32:
            word_data_offset, = struct.unpack(
                "!I", self._content[self._offset:self._offset + 4])
            self._offset += 4
        elif self._index_offset_bits == 64:
            word_data_offset, = struct.unpack(
                "!I", self._content[self._offset:self._offset + 8])
            self._offset += 8
        else:
            raise ValueError

        # Read word_data_size => always 32 bits
        word_data_size = 0
        word_data_size, = struct.unpack(
            "!I", self._content[self._offset:self._offset + 4])
        self._offset += 4

        # Increase index
        self._index += 1

        return (word_str, word_data_offset, word_data_size, self._index)

    def get_index_by_num(self, number):
        """Get index infomation of a specified entry in .idx file by origin index.
        May raise IndexError if number is out of range.

        Arguments:
        - `number`: the origin index of the entry in .idx file
        Return:
        A tuple in form of (word_str, word_data_offset, word_data_size)
        """
        if number >= len(self._index_idx):
            raise IndexError("Index out of range! Acessing the {:d} index but totally {:d}".format(
                number, len(self._index_idx)))
        return self._index_idx[number]

    def get_index_by_word(self, word_str):
        """Get index infomation of a specified word entry.

        Arguments:
        - `word_str`: name of word entry.
        Return:
        Index infomation corresponding to the specified word if exists, otherwise False.
        The index infomation returned is a list of tuples, in form of [(word_data_offset, word_data_size) ...]
        """
        if word_str not in self._word_idx:
            return []
        numbers = self._word_idx[word_str]
        indexes = []
        for number in numbers:
            indexes.append(self._index_idx[number][1:])
        return indexes

    def get_all_words(self):
        """Get all words in an dictionary

        Return:
        A set of words
        """

        return {data[0] for data in self._index_idx}


class DictFileReader(object):
    """Read the .dict file, store the data in memory for querying.
    """

    def __init__(self, filename, dict_ifo, dict_index, compressed=False):
        """Constructor.

        Arguments:
        - `filename`: filename of .dict file.
        - `dict_ifo`: IfoFileReader object.
        - `dict_index`: IdxFileReader object.
        """
        self._dict_ifo = dict_ifo
        self._dict_index = dict_index
        self._compressed = compressed
        self._offset = 0
        if self._compressed:
            with gzip.open(filename, "rb") as dict_file:
                self._dict_file = dict_file.read()
        else:
            with open(filename, "rb") as dict_file:
                self._dict_file = dict_file.read()

    def get_dict_by_word(self, word):
        """Get the word's dictionary data by it's name.

        Arguments:
        - `word`: word name.
        Return:
        The specified word's dictionary data, in form of dict as below:
        {type_identifier: infomation, ...}
        in which type_identifier can be any character in "mlgtxykwhnrWP".
        """
        result = []
        indexes = self._dict_index.get_index_by_word(word)
        if not indexes:
            return result
        # sametypesequence = m => same type_identifier = m
        sametypesequence = self._dict_ifo.get_ifo("sametypesequence")
        for index in indexes:
            self._offset = index[0]
            size = index[1]
            if sametypesequence:
                result.append(self._get_entry_sametypesequence(size))
            else:
                result.append(self._get_entry(size))
        return result

    def get_dict_by_index(self, index):
        """Get the word's dictionary data by it's index infomation.

        Arguments:
        - `index`: index of a word entrt in .idx file.'
        Return:
        The specified word's dictionary data, in form of dict as below:
        {type_identifier: infomation, ...}
        in which type_identifier can be any character in "mlgtxykwhnrWP".
        """
        word, offset, size = self._dict_index.get_index_by_num(index)
        self._offset = offset
        sametypesequence = self._dict_ifo.get_ifo("sametypesequence")
        if sametypesequence:
            return self._get_entry_sametypesequence(size)
        else:
            return self._get_entry(size)

    def _get_entry(self, size):
        result = {}
        read_size = 0
        start_offset = self._offset
        while read_size < size:
            type_identifier = struct.unpack("!c")
            if type_identifier in "mlgtxykwhnr":
                result[type_identifier] = self._get_entry_field_null_trail()
            else:
                result[type_identifier] = self._get_entry_field_size()
            read_size = self._offset - start_offset
        return result

    def _get_entry_sametypesequence(self, size):
        start_offset = self._offset
        result = {}
        sametypesequence = self._dict_ifo.get_ifo("sametypesequence")
        for k in range(0, len(sametypesequence)):
            if sametypesequence[k] in "mlgtxykwhnr":
                if k == len(sametypesequence) - 1:
                    result[sametypesequence[k]] = self._get_entry_field_size(
                        size - (self._offset - start_offset))
                else:
                    result[sametypesequence[k]
                           ] = self._get_entry_field_null_trail()
            elif sametypesequence[k] in "WP":
                if k == len(sametypesequence) - 1:
                    result[sametypesequence[k]] = self._get_entry_field_size(
                        size - (self._offset - start_offset))
                else:
                    result[sametypesequence[k]] = self._get_entry_field_size()
        return result

    def _get_entry_field_null_trail(self):
        end = self._dict_file.find(b'\x00', self._offset)
        result = self._dict_file[self._offset:end]
        self._offset = end + 1
        return result

    def _get_entry_field_size(self, size=None):
        if size is None:
            size = struct.unpack("!I", self._dict_file[
                                 self._offset:self._offset + 4])
            self._offset += 4
        result = self._dict_file[
            self._offset:self._offset + size]
        self._offset += size
        return result



from dataclasses import dataclass, field
from datetime import datetime
from time import time
from typing import List

from dataclasses_json import dataclass_json, LetterCase


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Translation:
    translation: str
    part_of_speech: str = ''
    transcript: List[str] = field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class WordMetadata:
    foreign_word: str
    test: str = ''
    translations: List[Translation] = field(default_factory=list)
    usage: List[str] = field(default_factory=set)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class SkovorodaDict:
    name: str               # "Petra's dict 2"
    description: str        # "Useful Hungarian words",
    language: str           # "Hungarian",
    date_created: str  = datetime.now().strftime('%d-%m-%Y %H:%S')  # "28-05-2020 15:15",
    words: List[WordMetadata] = field(default_factory=list)


# MARK: This models represent Kindle DB structure
@dataclass_json
@dataclass
class Word:
    id: str
    word: str
    stem: str
    lang: str
    timestamp: int = time()
    usage: List[str] = field(default_factory=list)

    @staticmethod
    def from_db_tupple(db_tupple):
        return Word(id=db_tupple[0], word=db_tupple[1],
                    stem=db_tupple[2], lang=db_tupple[3],
                    timestamp=db_tupple[5])


@dataclass_json
@dataclass
class Lookup:
    id: str
    word_key: str
    book_key: str
    dict_key: str
    pos: int
    usage: str
    timestamp: int


# !/usr/bin/env python
from os import path
from setuptools import setup, find_packages

this_directory = path.abspath(path.dirname(__file__))

with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='bingen',
      version='0.1.3',
      description='German to Ukrainian words translator',
      url='https://github.com/adyachok/bingen',
      author='Andras Gyacsok',
      author_email='atti.dyachok@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
            'build==0.7.0',
            'certifi==2021.10.8',
            'chardet==3.0.4',
            'charset-normalizer==2.0.12',
            'dataclasses-json==0.5.7',
            'googletrans==4.0.0rc1',
            'h11==0.9.0',
            'h2==3.2.0',
            'hpack==3.0.0',
            'hstspreload==2021.12.1',
            'httpcore==0.9.1',
            'httpx',
            'hyperframe==5.2.0',
            'idna==3.3',
            'marshmallow==3.15.0',
            'marshmallow-enum==1.5.1',
            'mypy-extensions==0.4.3',
            'packaging==21.3',
            'pep517==0.12.0',
            'pyparsing==3.0.8',
            'rfc3986==1.5.0',
            'sniffio==1.2.0',
            'tomli==2.0.1',
            'typing-inspect==0.7.1',
            'typing_extensions==4.2.0',
            'urllib3==1.26.9'
      ],
      long_description_content_type='text/markdown',
      long_description=long_description,
      package_data = {'bingen': ['stardict/de_dict/*']},
      zip_safe=False)


import re
from typing import List

from googletrans import Translator

from bingen.models import Translation, WordMetadata, SkovorodaDict
from bingen.stardict.processor import build_de_dict


def prepare_word_test(word:WordMetadata):
    test_str = "".join([tr.translation.strip() for tr in word.translations])
    foreign_word = word.foreign_word
    # Removes foreign word occurrence in translation
    test_str = test_str.replace('\n', ' ')
    test_str = test_str.replace(foreign_word, '~')
    # Removes transcript in translation
    match = re.search('\[.*\]', test_str)
    if match:
        test_str_beginning = test_str[:match.start()]
        test_str_ending = test_str[match.end()-1:]
        test_str = test_str_beginning + test_str_ending
    word.test = test_str[:251]


def translate(tokens: List[str]) -> SkovorodaDict:
    dicts = build_de_dict()
    not_translated_counter = 0
    sk_dict = SkovorodaDict(name='Bleib in Gottes Liebe Kap.4',
                            description='Bleib in Gottes Liebe. Kapitel 4. '
                                        'Autorität respektieren.',
                            language='German')
    for token in tokens:
        translation = dicts[0].get_dict_by_word(token)
        if translation:

            translation = translation[0].get('x')
            translation = str(translation, encoding='utf8')
            result = re.search('<k>\w.*</k>', translation)
            start, end = result.span()
            token = translation[start: end].strip('<k>').strip('</k>')
            translation = translation[end:]
            translation = translation.strip()
            translation = _translate_ua(translation)
            translation = Translation(translation=translation)
            token = WordMetadata(foreign_word=token)
            token.translations.append(translation)
            prepare_word_test(token)
            sk_dict.words.append(token)
        else:
            not_translated_counter += 1
    return sk_dict, not_translated_counter


def _translate_ua(text):
    translator = Translator()
    translation =  translator.translate(text, dest='uk', src='ru')
    text = translation.text
    if 'посл.' in text:
        text = text.replace('посл.', 'присл.')
    if 'кому-л.' in text:
        text = text.replace('кому-л.', 'кому-н.')
    if 'високий.' in text:
        text = text.replace('високий.', 'піднес.')
    return text


if __name__ == '__main__':
    tokens = ['Liebe']
    sk_dict, not_translated_counter = translate(tokens)
    print(f'Not translated: {not_translated_counter}')
    print(sk_dict.to_dict())



# Taken from: https://github.com/solominh/python-stardict
import logging
import os

from bingen.stardict.model import IdxFileReader, IfoFileReader, DictFileReader


logger = logging.basicConfig()


class Dictionary:

    def __init__(self, ifo_path, idx_path, dict_path):
        self._ifo_path = ifo_path
        self._idx_path = idx_path
        self._dict_path = dict_path
        self.dict_reader = self.init_dict_reader()

    def init_dict_reader(self):
        ifo_reader = IfoFileReader(self._ifo_path)
        idx_reader = IdxFileReader(self._idx_path,
                                   compressed=False,
                                   index_offset_bits=32)
        return DictFileReader(self._dict_path,
                              dict_ifo=ifo_reader,
                              dict_index=idx_reader,
                              compressed=True)


class DictionariesManager:

    @staticmethod
    def get_dict_files(dicts_path):
        dicts = []
        for root, dirs, files in os.walk(dicts_path):
            dict_files = {}
            for f in files:
                if f.endswith('dict.dz'):
                    dict_files['dict_path'] = os.path.join(root, f)
                elif f.endswith('.idx'):
                    dict_files['idx_path'] = os.path.join(root, f)
                elif f.endswith('.ifo'):
                    dict_files['ifo_path'] = os.path.join(root, f)
                else:
                    file_path = os.path.join(root, f)
                    # logger.warn(f'Found file {file_path}')
            dicts.append(dict_files)
        return dicts

    @staticmethod
    def build_dicts(dict_files):
        dict_files = [d for d in dict_files if len(d) == 3]
        dicts = [Dictionary(ifo_path=d.get('ifo_path'),
                            idx_path=d.get('idx_path'),
                            dict_path=d.get('dict_path')) for d in dict_files]
        return dicts

    @staticmethod
    def init_dicts(dicts):
        return [d.init_dict_reader() for d in dicts]


def build(dicts_path):
    dicts = DictionariesManager.get_dict_files(dicts_path)
    dicts = DictionariesManager.build_dicts(dicts)
    return DictionariesManager.init_dicts(dicts)


def build_de_dict():
    root_path = os.path.dirname(__file__)
    dicts_path = os.path.join(root_path, 'de_dict')
    return build(dicts_path)


if __name__ == '__main__':
    ROOT_PATH = os.path.dirname(__file__)
    DICTS_PATH = os.path.join(ROOT_PATH, 'stardict')
    dicts = build(DICTS_PATH)
    print(dicts[0].get_dict_by_word('Hello'))